from typing import Any

from ....data_schemas.v1.uoms import UnitType, Uom, UomConversionRequest, UomConversionResponse
from ..api import IDEXApi


class Uoms:
    def __init__(self, api: IDEXApi) -> None:
        self._api = api

    def get_uoms(self) -> list[Uom]:
        """
        Get `Uom` objects.

        Returns
        -------
        list[Uom]
            The `Uom` objects.

        """
        data = self._api.get(url="UnitOfMeasure/Uoms")
        return [Uom.model_validate(row) for row in data.json()]

    def get_unit_types(self) -> list[UnitType]:
        """
        Get `UnitType` objects.

        Returns
        -------
        list[UnitType]
            The `UnitType` objects.

        """
        data = self._api.get(url="UnitOfMeasure/UnitTypes")
        return [UnitType.model_validate(row) for row in data.json()]

    def convert_scalars(
        self,
        from_uom: str,
        to_uom: str,
        values: list[Any],
        include_source_values_in_response: bool = False,
    ) -> UomConversionResponse:
        """
        Convert scalar values from one unit of measure to another.

        Parameters
        ----------
        from_uom : str
            The source unit of measure.
        to_uom : str
            The target unit of measure.
        values : list[Any]
            The values to convert.

        Returns
        -------
        list[float]
            The converted values.

        """
        payload = UomConversionRequest.model_construct(
            source_uom_id=from_uom,
            target_uom_id=to_uom,
            values=values,
            include_source_values_in_response=include_source_values_in_response,
        )
        data = self._api.post(
            url="UnitOfMeasure/Convert",
            data=payload.model_dump_json(by_alias=True),
        )
        return UomConversionResponse.model_validate_json(data.text)



import logging
from typing import Any, Optional, Union, overload

from ....data_schemas.v1.assets import Reel
from ....data_schemas.v1.events import JobHistory, Maintenance, ScheduledJob
from ..api import IDEXApi


class Reels:
    def __init__(self, api: IDEXApi) -> None:
        self._api = api

    @overload
    def get(self, *, id: str) -> Reel: ...
    @overload
    def get(
        self,
        *,
        filter: Optional[str] = None,
        select: Optional[list[str]] = None,
        top: Optional[int] = None,
        skip: Optional[int] = None,
        order_by: Optional[str] = None,
    ) -> list[Reel]: ...

    def get(
        self,
        id: Optional[str] = None,
        filter: Optional[str] = None,
        select: Optional[list[str]] = None,
        top: Optional[int] = None,
        skip: Optional[int] = None,
        order_by: Optional[str] = None,
    ) -> Union[Reel, list[Reel]]:
        """
        Get `Reel` object(s).

        Parameters
        ----------
        id : Optional[str]
            Reel to retrieve.
        filter : Optional[str]
            OData filter string.
        select : list[str] | None
            Provide a list of columns to retrieve from output.
        top : Optional[int]
            Limit the number of results returned.
        skip : Optional[int]
            Skip the first N results.
        order_by : Optional[str]
            Order the results by columns.

        Returns
        -------
        Union[Reel, list[Reel]]
            The `Reel` object(s).

        """
        if id is not None:
            logging.debug(f"Getting Reel with ID: {id}")
            data = self._api.get(url=f"Reels/{id}")
            return Reel.model_validate(data.json())

        params: dict[str, Any] = {}
        if filter is not None:
            params["$filter"] = filter
        if select is not None:
            select = self._check_select(select)
            params["$select"] = ",".join(select)
        if top is not None:
            params["$top"] = top
        if skip is not None:
            params["$skip"] = skip
        if order_by is not None:
            params["$orderby"] = order_by

        data = self._api.get(url="Reels", params=params)

        if data.status_code == 204:
            return []

        return [Reel.model_validate(row) for row in data.json()]

    def _check_select(self, select: list[str]) -> list[str]:
        important_fields = ["id"]
        for field in important_fields:
            if field not in select:
                select.append(field)
        return select

    def get_maintenances(self, reel: Optional[Reel] = None, reel_id: Optional[str] = None) -> list[Maintenance]:
        if reel is not None:
            id = reel.id
        elif reel_id is not None:
            id = reel_id
        else:
            raise ValueError("Invalid input. Must provide either a Reel object or a reel_id.")

        logging.debug(f"Getting Maintenances for Reel with ID: {id}")
        data = self._api.get(url=f"Reels/{id}/Maintenances")
        if data.status_code == 204:
            return []

        return [Maintenance.model_validate(row) for row in data.json()]

    def get_job_history(self, reel: Optional[Reel] = None, reel_id: Optional[str] = None) -> list[JobHistory]:
        if reel is not None:
            id = reel.id
        elif reel_id is not None:
            id = reel_id
        else:
            raise ValueError("Invalid input. Must provide either a Reel object or a reel_id.")

        logging.debug(f"Getting JobHistory for Reel with ID: {id}")
        data = self._api.get(url=f"Reels/{id}/JobHistory")
        if data.status_code == 204:
            return []

        return [JobHistory.model_validate(row) for row in data.json()]

    def get_scheduled_jobs(self, reel: Optional[Reel] = None, reel_id: Optional[str] = None) -> list[ScheduledJob]:
        if reel is not None:
            id = reel.id
        elif reel_id is not None:
            id = reel_id
        else:
            raise ValueError("Invalid input. Must provide either a Reel object or a reel_id.")

        logging.debug(f"Getting ScheduledJobs for Reel with ID: {id}")
        data = self._api.get(url=f"Reels/{id}/ScheduledJobs")
        if data.status_code == 204:
            return []

        return [ScheduledJob.model_validate(row) for row in data.json()]



from ....data_schemas.v1.change_log import ChangeLog
from ..api import IDEXApi


class ChangeLogs:
    def __init__(self, api: IDEXApi) -> None:
        self._api = api

    def get_max_anchor(self) -> int:
        """
        Get change log max anchor.

        Returns
        -------
        int
            The maximum anchor value.

        """
        return self._api.get(url="ChangeLog/MaxAnchor").json()

    def get_changelogs(self, anchor_larger_than: int) -> list[ChangeLog]:
        """
        Get `ChangeLog` objects.

        Parameters
        ----------
        anchor_larger_than : int
            The anchor value for which returned `ChangeLog` objects should have a larger anchor value.

        """
        data = self._api.get(url=f"ChangeLog/Anchor/{anchor_larger_than}")

        return [ChangeLog.model_validate(d) for d in data.json()]



"""Some helpers for authentication purposes."""

import logging
from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import Any, Optional, Protocol
from urllib.parse import urljoin

import requests

from ...data_schemas.v1.auth import (
    AuthenticateRequest,
    AuthenticateResponse,
    RefreshTokenRequest,
    RefreshTokenResponse,
)


class IDEXAuth(Protocol):
    # Add post-init for parsing base url

    @property
    def base_url(self) -> str: ...

    def get_auth_header(self, **kwargs: Any) -> dict[str, str]: ...


@dataclass
class JwtAuth:
    base_url: str
    username: str
    password: str
    auth: Optional[AuthenticateResponse] = field(default=None, init=False, repr=False)

    @property
    def auth_request_payload(self) -> str:
        return AuthenticateRequest(
            username=self.username,
            password=self.password,
        ).model_dump_json()

    @property
    def refresh_request_payload(self) -> str:
        assert self.auth is not None
        return RefreshTokenRequest(
            refresh_token=self.auth.refresh_token,
        ).model_dump_json()

    def _acquire_token(self, base_url: str) -> AuthenticateResponse:
        """For acquiring a new token."""
        response = requests.post(
            url=urljoin(base_url, "Access/Authenticate"),
            headers={"Content-Type": "application/json"},
            data=self.auth_request_payload,
        )
        response.raise_for_status()
        logging.info("Received new token.")
        return AuthenticateResponse.model_validate_json(response.text)

    def _refresh_token(self, base_url: str) -> RefreshTokenResponse:
        """For refreshing token."""
        response = requests.post(
            url=urljoin(base_url, "Access/RefreshToken"),
            headers={"Content-Type": "application/json"},
            data=self.refresh_request_payload,
        )
        response.raise_for_status()
        logging.info("Refreshed token.")
        return RefreshTokenResponse.model_validate_json(response.text)

    def get_auth_header(self, **kwargs: Any) -> dict[str, str]:
        try:
            base_url = kwargs["base_url"]
        except KeyError:
            raise ValueError("The `base_url` keyword must be provided.")

        if self.auth is None:
            self.auth = self._acquire_token(base_url=base_url)

        elif self.auth.refresh_token_expire <= datetime.now(tz=timezone.utc):
            self.auth = self._acquire_token(base_url=base_url)

        elif self.auth.token_expire <= datetime.now(tz=timezone.utc):
            resp = self._refresh_token(base_url=base_url)
            self.auth.token = resp.token
            self.auth.token_expire = resp.token_expire

        return {"Authorization": self.auth.token}


@dataclass(frozen=True)
class ApiKeyAuth:
    base_url: str
    x_api_key: str

    def get_auth_header(self, **kwargs: Any) -> dict[str, str]:
        return {"X-API-KEY": self.x_api_key}



from datetime import datetime
from .base import IDEX


class AuthenticateRequest(IDEX):
    """Describes a request to authenticate a user."""

    username: str
    password: str


class AuthenticateResponse(IDEX):
    """Describes a response to an authentication request."""

    id: str
    first_name: str
    last_name: str
    username: str
    token: str
    token_expire: datetime
    refresh_token: str
    refresh_token_expire: datetime


class RefreshTokenRequest(IDEX):
    """Describes a request to refresh a user token."""

    refresh_token: str


class RefreshTokenResponse(IDEX):
    """Describes a response to a token refresh request."""

    token: str
    token_expire: datetime



import logging
from datetime import datetime
from typing import Optional, overload

from ....data_schemas.v1.assets import Log
from ....data_schemas.v1.channel_data import (
    Channel,
    ChannelDataRangeRequest,
    ChannelDataRequest,
    ChannelDataResponse,
    ChannelRange,
)
from ....data_schemas.v1.events import Run
from ..api import IDEXApi
from .text_utils import url_encode_id


class Channels:
    def __init__(self, api: IDEXApi) -> None:
        self._api = api

    @overload
    def get(self, *, log: Log, include_soft_delete: Optional[bool] = False) -> list[Channel]: ...
    @overload
    def get(self, *, log_id: str, include_soft_delete: Optional[bool] = False) -> list[Channel]: ...
    @overload
    def get(self, *, run: Run, include_soft_delete: Optional[bool] = False) -> list[Channel]: ...
    @overload
    def get(self, *, run_id: str, include_soft_delete: Optional[bool] = False) -> list[Channel]: ...

    def get(
        self,
        *,
        log: Optional[Log] = None,
        log_id: Optional[str] = None,
        run: Optional[Run] = None,
        run_id: Optional[str] = None,
        include_soft_delete: Optional[bool] = False,
    ) -> list[Channel]:
        """
        Get `Channel` object(s).

        Parameters
        ----------
        log : Optional[Log]
            Log object to get Channel for.
        log_id : Optional[str]
            Log ID to get Channel for.
        run : Optional[Run]
            Run object to get Channel for.
        run_id : Optional[str]
            Run ID to get Channel for.
        include_soft_delete : Optional[bool] = False
            Include soft deleted records.

        Returns
        -------
        list[Channel]
            The `Channel` object(s).

        """
        log_id_, run_id_ = None, None

        if log is not None or log_id is not None:
            log_id_ = log_id if log is None else log.id
            assert log_id_ is not None
            logging.debug(f"Getting Channels for Log with ID: {id}")
            log_id_ = url_encode_id(log_id_)
            data = self._api.get(url=f"Logs/{log_id_}/Channels")

        elif run is not None or run_id is not None:
            run_id_ = run_id if run is None else run.id
            assert run_id_ is not None
            logging.debug(f"Getting Channels for Run with ID: {id}")
            run_id_ = url_encode_id(run_id_)
            data = self._api.get(url=f"Runs/{run_id_}/Channels")

        else:
            raise TypeError("Either `log`, `log_id`, `run` or `run_id` must be provided.")

        if data.status_code == 204:
            return []

        records = [Channel.model_validate(row) for row in data.json()]

        if log_id_ is not None:
            for record in records:
                record.log_id = log_id_

        if run_id_ is not None:
            for record in records:
                record.run_id = run_id_

        if include_soft_delete:
            return records

        return [rec for rec in records if rec.deleted_date is None]

    @overload
    def get_available_ranges(self, *, channels: list[Channel]) -> list[ChannelRange]: ...
    @overload
    def get_available_ranges(self, *, channel_ids: list[str]) -> list[ChannelRange]: ...

    def get_available_ranges(
        self,
        *,
        channels: Optional[list[Channel]] = None,
        channel_ids: Optional[list[str]] = None,
    ) -> list[ChannelRange]:
        """
        Get `ChannelRange` object(s).

        Parameters
        ----------
        channels : Optional[list[Channel]]
            Channel objects to get Ranges for.
        channel_ids : Optional[list[str]]
            Channel IDs to get Ranges for.

        Returns
        -------
        list[ChannelRange]
            The `ChannelRange` object(s).

        """
        if channels is not None:
            ids = [channel.id for channel in channels]
        elif channel_ids is not None:
            ids = channel_ids
        else:
            raise TypeError("Either `channels` or `channel_ids` must be provided.")

        logging.debug(f"Getting available ranges for Channels with IDs: {', '.join(ids)}.")

        data = self._api.post(url="ChannelData/AvailableRanges", json=ids)

        if data.status_code == 204:
            return []

        return [ChannelRange.model_validate(row) for row in data.json()]

    @overload
    def get_data_range(
        self,
        *,
        channels: list[Channel],
        start_time: datetime,
        end_time: datetime,
        limit: int,
        ignore_unknown_ids: bool = True,
        include_outside_pts: bool = True,
    ) -> list[ChannelDataResponse]: ...
    @overload
    def get_data_range(
        self,
        *,
        channel_ids: list[str],
        start_time: datetime,
        end_time: datetime,
        limit: int,
        ignore_unknown_ids: bool = True,
        include_outside_pts: bool = True,
    ) -> list[ChannelDataResponse]: ...

    def get_data_range(
        self,
        *,
        start_time: datetime,
        end_time: datetime,
        limit: int,
        ignore_unknown_ids: bool = True,
        include_outside_pts: bool = True,
        channels: Optional[list[Channel]] = None,
        channel_ids: Optional[list[str]] = None,
    ) -> list[ChannelDataResponse]:
        """
        Get `ChannelDataResponse` object(s).

        Parameters
        ----------
        channels : Optional[list[Channel]]
            Channel objects to get Ranges for.
        channel_ids : Optional[list[str]]
            Channel IDs to get Ranges for.
        start_time : datetime
            The start time for channel data.
        end_time : datetime
            The end time for channel data.
        limit : int
            The limit of datapoints to retrieve per channel.
        ignore_unknown_ids : bool = True
            Ignore unknown channel IDs.
        include_outside_pts : bool = True
            Include outside points.


        Returns
        -------
        list[ChannelDataResponse]
            The `ChannelDataResponse` object(s).

        """
        payload = ChannelDataRequest(
            ids=[channel.id for channel in channels] if channels is not None else channel_ids,  # type: ignore
            start=start_time,
            end=end_time,
            limit=limit,
            ignore_unknown_ids=ignore_unknown_ids,
            include_outside_points=include_outside_pts,
        )

        data = self._api.post(url="ChannelData/GetDataRange", data=payload.model_dump_json(by_alias=True))

        if data.status_code == 204:
            return []

        return [ChannelDataResponse.model_validate(row) for row in data.json()]

    def get_channel_datapoints(
        self,
        *,
        channel_ranges: list[ChannelRange],
        limit: int,
        ignore_unknown_ids: bool = True,
        include_outside_pts: bool = True,
    ) -> bytes:
        """
        Get `ChannelDataResponse` object(s).

        Parameters
        ----------
        channels : list[ChannelRange]
            Channel objects to get Ranges for.
        limit : int
            The limit of datapoints to retrieve per channel.
        ignore_unknown_ids : bool = True
            Ignore unknown channel IDs.
        include_outside_pts : bool = True
            Include outside points.


        Returns
        -------
        list[ChannelDataResponse]
            The `ChannelDataResponse` object(s).

        """
        payload = ChannelDataRangeRequest(
            channels=channel_ranges,
            limit=limit,
            ignore_unknown_ids=ignore_unknown_ids,
            include_outside_points=include_outside_pts,
        )

        data = self._api.post(
            url="ChannelData/GetChannelDataRange",
            data=payload.model_dump_json(by_alias=True),
        )

        if data.status_code == 204:
            return b""

        return data.content

    def get_compressed_channel_datapoints(
        self,
        *,
        channel_ranges: list[ChannelRange],
        limit: int,
        ignore_unknown_ids: bool = True,
        include_outside_pts: bool = True,
    ) -> bytes:
        """
        Get `ChannelDataResponse` object(s).

        Response is compressed using Brotli compression and MessagePack.

        NOTE: Currently responds with an undocumented 406 error.

        Parameters
        ----------
        channels : list[ChannelRange]
            Channel objects to get Ranges for.
        limit : int
            The limit of datapoints to retrieve per channel.
        ignore_unknown_ids : bool = True
            Ignore unknown channel IDs.
        include_outside_pts : bool = True
            Include outside points.


        Returns
        -------
        list[ChannelDataResponse]
            The `ChannelDataResponse` object(s).

        """
        payload = ChannelDataRangeRequest(
            channels=channel_ranges,
            limit=limit,
            ignore_unknown_ids=ignore_unknown_ids,
            include_outside_points=include_outside_pts,
        )

        data = self._api.post(
            url="ChannelData/GetCompressedChannelDataRange",
            data=payload.model_dump_json(by_alias=True),
            headers={"Accept": "application/octet-stream"},
        )

        if data.status_code == 204:
            return b""

        return data.content



from ....data_schemas.v1.assets import Coil
from ..api import IDEXApi


class Coils:
    def __init__(self, api: IDEXApi) -> None:
        self._api = api

    def get(self, manufacturer: str) -> list[Coil]:
        """
        Get `Coil` object(s).

        Parameters
        ----------
        manufacturer : str
            Manufacturer to retrieve `Coil` objects for.

        Returns
        -------
        list[Coil]
            The `Coil` object(s).

        """
        data = self._api.get(url=f"Coils/{manufacturer}")

        return [Coil.model_validate(d) for d in data.json()]



from datetime import datetime
from typing import Optional

from .base import IDEX


class ChangeLog(IDEX):
    id: str
    anchor: int
    type: Optional[str]
    source: Optional[str]
    created_date: Optional[datetime]



from typing import Optional

from requests import Session

from .api import IDEXApi
from .auth import IDEXAuth
from .modules import (
    ChangeLogs,
    Channels,
    Coils,
    Customers,
    Fields,
    Installations,
    Logs,
    Reels,
    Runs,
    Soe,
    Surveys,
    Tests,
    UnitActiveWellbores,
    Units,
    Uoms,
    WellboreHistories,
    WellboreLiveStatuses,
    Wellbores,
    Wells,
)


class IDEXClient:
    def __init__(self, *, auth: IDEXAuth, session: Optional[Session] = None) -> None:
        if session is None:
            session = Session()

        self._api = IDEXApi(auth=auth, session=session)

        self.change_logs = ChangeLogs(api=self._api)
        self.channels = Channels(api=self._api)
        self.coils = Coils(api=self._api)
        self.customers = Customers(api=self._api)
        self.fields = Fields(api=self._api)
        self.installations = Installations(api=self._api)
        self.logs = Logs(api=self._api)  # Unit ID seems to not work.
        self.reels = Reels(api=self._api)
        self.runs = Runs(api=self._api)
        self.soe = Soe(api=self._api)
        self.surveys = Surveys(api=self._api)
        self.tests = Tests(api=self._api)
        self.unit_active_wellbores = UnitActiveWellbores(api=self._api)
        self.units = Units(api=self._api)
        self.uoms = Uoms(api=self._api)
        self.wellbore_histories = WellboreHistories(api=self._api)
        self.wellbore_live_statuses = WellboreLiveStatuses(api=self._api)
        self.wells = Wells(api=self._api)
        self.wellbores = Wellbores(api=self._api)



import logging
from typing import Optional, overload

from ....data_schemas.v1.assets import Wellbore
from ....data_schemas.v1.events import (
    SoeActivity,
    SoeChemicalMeasurement,
    SoeJob,
    SoeTask,
)
from ..api import IDEXApi


class Soe:
    def __init__(self, api: IDEXApi) -> None:
        self._api = api

    def _get_jobs(self, wellbore_id: str) -> list[SoeJob]:
        data = self._api.get(url=f"SequenceOfEvents/{wellbore_id}/Jobs")
        if data.status_code == 204:
            return []

        return [SoeJob.model_validate(row) for row in data.json()]

    def _get_tasks(self, wellbore_id: str, job_id: str) -> list[SoeTask]:
        data = self._api.get(url=f"SequenceOfEvents/{wellbore_id}/Jobs/{job_id}/Tasks")
        if data.status_code == 204:
            return []  # Empty response

        tasks = [SoeTask.model_validate(row) for row in data.json()]
        for task in tasks:  # Enriching with wellbore_id, not part of API payload
            task.wellbore_id = wellbore_id
        return tasks

    def _get_chemical_measurements(self, wellbore_id: str, job_id: str) -> list[SoeChemicalMeasurement]:
        data = self._api.get(url=f"SequenceOfEvents/{wellbore_id}/Jobs/{job_id}/ChemicalMeasurements")
        if data.status_code == 204:
            return []

        measurements = [SoeChemicalMeasurement.model_validate(row) for row in data.json()]
        for measurement in measurements:
            measurement.wellbore_id = wellbore_id
        return measurements

    def _get_activities(self, wellbore_id: str, job_id: str, task_id: str) -> list[SoeActivity]:
        data = self._api.get(url=f"SequenceOfEvents/{wellbore_id}/Jobs/{job_id}/Tasks/{task_id}/Activities")
        if data.status_code == 204:
            return []

        return [SoeActivity.model_validate(row) for row in data.json()]

    @overload
    def get_jobs(self, *, wellbore: Wellbore) -> list[SoeJob]: ...
    @overload
    def get_jobs(self, *, wellbore_id: str) -> list[SoeJob]: ...

    def get_jobs(self, *, wellbore: Optional[Wellbore] = None, wellbore_id: Optional[str] = None) -> list[SoeJob]:
        """
        Get `SoeJob` objects for a given Wellbore.

        Parameters
        ----------
        wellbore : Optional[Wellbore]
            The Wellbore object for which to retrieve all Jobs.
        wellbore_id : Optional[str]
            The ID of the Wellbore for which to retrieve all Jobs.

        Returns
        -------
        list[SoeJob]
            The `SoeJob` objects for the Wellbore.

        """
        if wellbore is not None:
            logging.debug(f"Getting Jobs for Wellbore with ID: {wellbore.id}")
            return self._get_jobs(wellbore_id=wellbore.id)

        if wellbore_id is not None:
            logging.debug(f"Getting Jobs for Wellbore with ID: {wellbore_id}")
            return self._get_jobs(wellbore_id=wellbore_id)  # type: ignore

        raise ValueError("Either wellbore or wellbore_id must be provided.")

    @overload
    def get_tasks(self, *, job: SoeJob) -> list[SoeTask]: ...
    @overload
    def get_tasks(self, *, wellbore_id: str, job_id: str) -> list[SoeTask]: ...

    def get_tasks(
        self, *, job: Optional[SoeJob] = None, wellbore_id: Optional[str] = None, job_id: Optional[str] = None
    ) -> list[SoeTask]:
        """
        Get `SoeTask` objects for a given SoE Job.

        Use either a SoeJob object or provide the wellbore_id and job_id explicitly.

        Parameters
        ----------
        job : Optional[SoeJob]
            The SoeJob object for which to retrieve all Tasks.
        wellbore_id : Optional[str]
            The ID of the Wellbore for which to retrieve all Tasks.
        job_id : Optional[str]
            The ID of the Job for which to retrieve all Tasks.

        Returns
        -------
        list[SoeTask]
            The `SoeTask` objects for the Job.

        """
        if job is not None:
            logging.debug(f"Getting Tasks for Job ID: {job.id}")
            return self._get_tasks(wellbore_id=job.wellbore_id, job_id=job.id)

        if wellbore_id is not None and job_id is not None:
            logging.debug(f"Getting Tasks for Job ID: {job_id}")
            return self._get_tasks(wellbore_id=wellbore_id, job_id=job_id)

        raise ValueError("Either job or wellbore_id and job_id must be provided.")

    @overload
    def get_chemical_measurements(self, *, job: SoeJob) -> list[SoeChemicalMeasurement]: ...
    @overload
    def get_chemical_measurements(self, *, wellbore_id: str, job_id: str) -> list[SoeChemicalMeasurement]: ...

    def get_chemical_measurements(
        self, *, job: Optional[SoeJob] = None, wellbore_id: Optional[str] = None, job_id: Optional[str] = None
    ) -> list[SoeChemicalMeasurement]:
        """
        Get `SoeChemicalMeasurement` objects for a given SoE Job.

        Use either a `SoeJob` object or provide the wellbore_id and job_id explicitly.

        Parameters
        ----------
        job : Optional[SoeJob]
            The SoeJob object for which to retrieve all Chemical Measurements.
        wellbore_id : Optional[str]
            The ID of the Wellbore for which to retrieve all Chemical Measurements.
        job_id : Optional[str]
            The ID of the Job for which to retrieve all Chemical Measurements.

        Returns
        -------
        list[SoeChemicalMeasurement]
            The `SoeChemicalMeasurement` objects for the Job.

        """
        if job is not None:
            logging.debug(f"Getting Chemical Measurements for Job ID: {job.id}")
            return self._get_chemical_measurements(wellbore_id=job.wellbore_id, job_id=job.id)
        if all(v is not None for v in [wellbore_id, job_id]):
            logging.debug(f"Getting Chemical Measurements for Job ID: {job_id}")
            return self._get_chemical_measurements(wellbore_id=wellbore_id, job_id=job_id)  # type: ignore

        raise ValueError("Either job or wellbore_id and job_id must be provided.")

    @overload
    def get_activities(self, *, task: SoeTask) -> list[SoeActivity]: ...
    @overload
    def get_activities(self, *, wellbore_id: str, job_id: str, task_id: str) -> list[SoeActivity]: ...

    def get_activities(
        self,
        *,
        task: Optional[SoeTask] = None,
        wellbore_id: Optional[str] = None,
        job_id: Optional[str] = None,
        task_id: Optional[str] = None,
    ) -> list[SoeActivity]:
        """
        Get `SoeActivity` objects for a given SoE Job Task.

        Use either a SoeTask object or provide the wellbore_id, job_id and task_id explicitly.

        Parameters
        ----------
        job : Optional[SoeJob]
            The SoeJob object for which to retrieve all Chemical Measurements.
        wellbore_id : Optional[str]
            The ID of the Wellbore for which to retrieve all Chemical Measurements.
        job_id : Optional[str]
            The ID of the Job for which to retrieve all Chemical Measurements.
        task_id : Optional[str]
            The ID of the Task for which to retrieve all Activities.

        Returns
        -------
        list[SoeActivity]
            The `SoeActivity` objects for the Task.

        """
        if task is not None:
            logging.debug(f"Getting Activities for Task ID: {task.id}")
            return self._get_activities(wellbore_id=task.wellbore_id, job_id=task.job_id, task_id=task.id)  # type: ignore
        if all(v is not None for v in [wellbore_id, job_id, task_id]):
            logging.debug(f"Getting Activities for Task ID: {task_id}")
            return self._get_activities(wellbore_id=wellbore_id, job_id=job_id, task_id=task_id)  # type: ignore

        raise ValueError("Provide either a `SoeTask` or a wellbore_id, job_id, and task_id combination.")



import logging
from typing import Optional, overload

from ....data_schemas.v1.assets import Wellbore, WellboreLiveStatus
from ..api import IDEXApi
from .text_utils import url_encode_id


class WellboreLiveStatuses:
    def __init__(self, api: IDEXApi) -> None:
        self._api = api

    @overload
    def get(self, *, include_soft_delete: Optional[bool] = False) -> list[WellboreLiveStatus]: ...
    @overload
    def get(self, *, wellbore: Wellbore, include_soft_delete: Optional[bool] = False) -> list[WellboreLiveStatus]: ...
    @overload
    def get(self, *, wellbore_id: str, include_soft_delete: Optional[bool] = False) -> list[WellboreLiveStatus]: ...

    def get(
        self,
        *,
        wellbore: Optional[Wellbore] = None,
        wellbore_id: Optional[str] = None,
        include_soft_delete: Optional[bool] = False,
    ) -> list[WellboreLiveStatus]:
        """
        Get `WellboreLiveStatus` object(s).

        Parameters
        ----------
        wellbore : Optional[Wellbore]
            Well object to get WellboreLiveStatus for.
        wellbore_id : Optional[str]
            Wellbore ID to get WellboreLiveStatus for.
        include_soft_delete : Optional[bool] = False
            Include soft deleted records.

        Returns
        -------
        list[WellboreLiveStatus]
            The `WellboreLiveStatus` object(s).

        """
        if wellbore is not None or wellbore_id is not None:
            id = wellbore_id if wellbore is None else wellbore.id
            assert id is not None
            logging.debug(f"Getting WellboreLiveStatus for Wellbore with ID: {id}")
            id = url_encode_id(id)
            data = self._api.get(url=f"WellboreLiveStatus/{id}")

        else:
            data = self._api.get(url="WellboreLiveStatus")

        if data.status_code == 204:
            return []

        records = [WellboreLiveStatus.model_validate(row) for row in data.json()]

        if include_soft_delete:
            return records

        return [rec for rec in records if rec.deleted_date is None]



import logging
from typing import Any, Optional, overload

from ....data_schemas.v1.assets import Unit, Wellbore
from ....data_schemas.v1.events import Run
from ..api import IDEXApi
from .text_utils import url_encode_id


class Runs:
    def __init__(self, api: IDEXApi) -> None:
        self._api = api

    @overload
    def get(self, *, unit: Unit, include_soft_delete: Optional[bool] = False) -> list[Run]: ...
    @overload
    def get(self, *, wellbore: Wellbore, include_soft_delete: Optional[bool] = False) -> list[Run]: ...
    @overload
    def get(self, *, unit_id: str, include_soft_delete: Optional[bool] = False) -> list[Run]: ...
    @overload
    def get(self, *, wellbore_id: str, include_soft_delete: Optional[bool] = False) -> list[Run]: ...
    @overload
    def get(
        self,
        *,
        filter: Optional[str] = None,
        select: Optional[list[str]] = None,
        top: Optional[int] = None,
        skip: Optional[int] = None,
        order_by: Optional[str] = None,
        include_soft_delete: Optional[bool] = False,
    ) -> list[Run]: ...

    def get(
        self,
        unit: Optional[Unit] = None,
        unit_id: Optional[str] = None,
        wellbore: Optional[Wellbore] = None,
        wellbore_id: Optional[str] = None,
        filter: Optional[str] = None,
        select: Optional[list[str]] = None,
        top: Optional[int] = None,
        skip: Optional[int] = None,
        order_by: Optional[str] = None,
        include_soft_delete: Optional[bool] = False,
    ) -> list[Run]:
        """
        Get `Run` objects.

        Parameters
        ----------
        unit : Optional[Unit]
            Unit object to get Runs for.
        unit_id : Optional[str]
            Unit ID to get Runs for.
        wellbore : Optional[Wellbore]
            Wellbore object to get Runs for.
        wellbore_id : Optional[str]
            Wellbore ID to get Runs for.
        filter : Optional[str]
            OData filter string.
        select : list[str] | None
            Provide a list of columns to retrieve from output.
        top : Optional[int]
            Limit the number of results returned.
        skip : Optional[int]
            Skip the first N results.
        order_by : Optional[str]
            Order the results by columns.
        include_soft_delete : Optional[bool] = False
            Include soft deleted records.

        Returns
        -------
        list[Run]
            The `Run` objects.

        """
        if unit is not None or unit_id is not None:
            id = unit_id if unit is None else unit.id
            assert id is not None
            logging.debug(f"Getting Runs for Unit with ID: {id}")
            id = url_encode_id(id)
            data = self._api.get(url=f"Units/{id}/Runs")

        elif wellbore is not None or wellbore_id is not None:
            id = wellbore_id if wellbore is None else wellbore.id
            assert id is not None
            logging.debug(f"Getting Runs for Wellbore with ID: {id}")
            id = url_encode_id(id)
            data = self._api.get(url=f"Wellbores/{id}/Runs")

        else:
            params: dict[str, Any] = {}
            if filter is not None:
                params["$filter"] = filter
            if select is not None:
                select = self._check_select(select)
                params["$select"] = ",".join(select)
            if top is not None:
                params["$top"] = top
            if skip is not None:
                params["$skip"] = skip
            if order_by is not None:
                params["$orderby"] = order_by

            data = self._api.get(url="Runs", params=params)

        if data.status_code == 204:
            logging.debug("No Runs found.")
            return []

        runs = [Run.model_validate(row) for row in data.json()]

        if include_soft_delete:
            return runs

        return [run for run in runs if run.deleted_date is None]

    def _check_select(self, select: list[str]) -> list[str]:
        important_fields = ["id"]
        for field in important_fields:
            if field not in select:
                select.append(field)
        return select



from typing import Optional

from pydantic import Field

from .base import IDEX


class UomInfo(IDEX):
    id: str


class Uom(UomInfo):
    name: Optional[str]
    description: Optional[str]
    is_base: Optional[bool]
    base_uom_id: Optional[str]


class UnitType(IDEX):
    id: str
    name: str
    member_uoms: list[UomInfo]


class UomConversionRequest(IDEX):
    source_uom_id: str
    target_uom_id: str
    values: list[float]
    include_source_values_in_response: bool = Field(default=False)


class UomConversionResponse(IDEX):
    source_uom_id: str
    target_uom_id: str
    converted_values: list[float]
    source_values: list[float]



import logging
from typing import Optional, Union, overload

from ....data_schemas.v1.assets import Log, Unit
from ..api import IDEXApi
from .text_utils import url_encode_id


class Logs:
    def __init__(self, api: IDEXApi) -> None:
        self._api = api

    @overload
    def get(self, *, id: str) -> Log: ...
    @overload
    def get(self, *, unit: Unit) -> list[Log]: ...
    @overload
    def get(self, *, unit_id: str) -> list[Log]: ...

    def get(
        self,
        *,
        id: Optional[str] = None,
        unit: Optional[Unit] = None,
        unit_id: Optional[str] = None,
    ) -> Union[Log, list[Log]]:
        """
        Get `Log` object(s).

        Parameters
        ----------
        id : Optional[str]
            Log to retrieve.
        unit : Optional[Unit]
            Unit object to get Logs for.
        unit_id : Optional[str]
            Unit ID to get Logs for.

        Returns
        -------
        Union[Log, list[Log]]
            The `Log` object(s).

        """
        if id is not None:
            logging.debug(f"Getting Wellbore with ID: {id}")
            data = self._api.get(url=f"Logs/{id}")
            return Log.model_validate(data.json())

        if unit is not None:
            logging.debug(f"Getting Logs for Unit with ID: {unit.id}")
            id = url_encode_id(unit.id)
        elif unit_id is not None:
            logging.debug(f"Getting Logs for Unit with ID: {unit_id}")
            id = url_encode_id(unit_id)
        else:
            raise ValueError("Either `unit` or `unit_id` must be provided.")

        data = self._api.get(url=f"Units/{id}/Logs")
        if data.status_code == 204:
            return []

        return [Log.model_validate(row) for row in data.json()]



from typing import Optional

from pydantic import Field

from ..base import IDEX, Approval, IDEXAudit, IDEXAuditLite, TimeDuration


class JobPlan(IDEX):
    """
    Describes a Job Plan.

    Endpoint allows for selecting return columns, need default Nones.
    """

    id: str
    name: Optional[str] = Field(default=None)
    well_name: Optional[str] = Field(default=None)
    well_id: Optional[str] = Field(default=None)
    wellbore_name: Optional[str] = Field(default=None)
    wellbore_id: Optional[str] = Field(default=None)
    field_name: Optional[str] = Field(default=None)
    phase: Optional[str] = Field(default=None)
    service_type: Optional[str] = Field(default=None)
    operator_name: Optional[str] = Field(default=None)
    status: Optional[str] = Field(default=None)
    project_auto_id: Optional[str] = Field(default=None)
    audit: IDEXAudit


class JobPlanRequest(IDEX):
    """For getting plans that match filter. All properties are optional."""

    well_name: Optional[str] = Field(default=None)
    well_id: Optional[str] = Field(default=None)
    wellbore_name: Optional[str] = Field(default=None)
    wellbore_id: Optional[str] = Field(default=None)
    field_name: Optional[str] = Field(default=None)
    creation_date: Optional[str] = Field(default=None)
    phase: Optional[str] = Field(default=None)
    service_type: Optional[str] = Field(default=None)


class _IDEXActivity:
    id: str
    name: Optional[str]
    type: Optional[str]
    order: int


class JobPlanTask(_IDEXActivity):
    approvals: list[Approval]


class Activity(_IDEXActivity):
    estimated_durations: list[TimeDuration]


class Step(_IDEXActivity): ...


class Document(IDEXAuditLite):
    """
    Describes a Document.

    Endpoint allows for selecting return columns, need default Nones.
    """

    id: str
    job_plan_id: Optional[str]
    job_plan_name: Optional[str] = Field(default=None)
    file_name: Optional[str] = Field(default=None)
    file_type: Optional[str] = Field(default=None)
    document_kind: Optional[str] = Field(default=None)
    asset: Optional[str] = Field(default=None)
    service_type: Optional[str] = Field(default=None)
    job_type: Optional[str] = Field(default=None)
    well_name: Optional[str] = Field(default=None)
    well_id: Optional[str] = Field(default=None)
    wellbore_name: Optional[str] = Field(default=None)
    wellbore_id: Optional[str] = Field(default=None)
    field_name: Optional[str] = Field(default=None)



from typing import Any, Optional

from ....data_schemas.v1.assets import Unit
from ..api import IDEXApi


class Units:
    def __init__(self, api: IDEXApi) -> None:
        self._api = api

    def get(
        self,
        filter: Optional[str] = None,
        select: Optional[list[str]] = None,
        top: Optional[int] = None,
        skip: Optional[int] = None,
        order_by: Optional[str] = None,
    ) -> list[Unit]:
        """
        Get `Unit` objects.

        Parameters
        ----------
        filter : Optional[str]
            OData filter string.
        select : list[str] | None
            Provide a list of columns to retrieve from output.
        top : Optional[int]
            Limit the number of results returned.
        skip : Optional[int]
            Skip the first N results.
        order_by : Optional[str]
            Order the results by columns.

        Returns
        -------
        list[Unit]
            The `Unit` objects.

        """
        params: dict[str, Any] = {}
        if filter is not None:
            params["$filter"] = filter
        if select is not None:
            select = self._check_select(select)
            params["$select"] = ",".join(select)
        if top is not None:
            params["$top"] = top
        if skip is not None:
            params["$skip"] = skip
        if order_by is not None:
            params["$orderby"] = order_by

        data = self._api.get(url="Units", params=params)

        if data.status_code == 204:
            return []

        return [Unit.model_validate(row) for row in data.json()]

    def _check_select(self, select: list[str]) -> list[str]:
        important_fields = ["id"]
        for field in important_fields:
            if field not in select:
                select.append(field)
        return select



from .change_logs import ChangeLogs
from .channels import Channels
from .coils import Coils
from .customers import Customers
from .fields import Fields
from .installations import Installations
from .logs import Logs
from .reels import Reels
from .runs import Runs
from .soe import Soe
from .surveys import Surveys
from .tests import Tests
from .unit_active_wellbore import UnitActiveWellbores
from .units import Units
from .uoms import Uoms
from .wellbore_history import WellboreHistories
from .wellbore_live_status import WellboreLiveStatuses
from .wellbores import Wellbores
from .wells import Wells

__all__ = [
    "ChangeLogs",
    "Channels",
    "Coils",
    "Customers",
    "Fields",
    "Installations",
    "Logs",
    "Reels",
    "Runs",
    "Soe",
    "Surveys",
    "Tests",
    "UnitActiveWellbores",
    "Units",
    "Uoms",
    "WellboreHistories",
    "WellboreLiveStatuses",
    "Wellbores",
    "Wells",
]



from datetime import datetime
from typing import Optional

from pydantic import Field

from ..base import IDEX, DoubleNullableUomValue, IDEXAuditLite


class SoeSensorValues(IDEX):
    """Describes a sensor value for a given Job."""

    name: Optional[str]
    global_name: Optional[str]
    uom: Optional[str]
    start_value: Optional[float]
    end_value: Optional[float]
    max_value: Optional[float]
    min_value: Optional[float]
    avg_value: Optional[float]


class SoeActivity(IDEXAuditLite):
    """Describes an Activity in a SoeTask."""

    id: str
    task_id: str
    name: Optional[str]
    type: Optional[str]
    start: Optional[datetime]
    end: Optional[datetime]
    duration: Optional[str]
    sensor_values: list[SoeSensorValues]
    comment: Optional[str]


class SoeTask(IDEXAuditLite):
    """Describes a SoeTask for a given Job."""

    id: str
    job_id: str
    name: Optional[str]
    conveyance_type: Optional[str]
    activities_count: int
    sequence_number: int
    state: Optional[str]
    start: Optional[datetime]
    end: Optional[datetime]
    running_down_length: DoubleNullableUomValue
    running_up_length: DoubleNullableUomValue
    running_total_length: DoubleNullableUomValue
    max_depth: DoubleNullableUomValue
    wellbore_id: Optional[str] = Field(default=None)  # Not part of the payload - added separately


class SoeChemicalMeasurement(IDEX):
    """Describes a chemical measurement for an Activity."""

    id: str
    job_id: str
    date: datetime
    type: Optional[str]
    value: float
    uom: Optional[str]
    comment: Optional[str]
    wellbore_id: Optional[str] = Field(default=None)  # Not part of the payload - added separately


class SoeJob(IDEXAuditLite):
    """Describes a Sequence of Events (Soe) Job."""

    id: str
    wellbore_id: str
    name: Optional[str]
    start: Optional[datetime]
    end: Optional[datetime]
    job_type_name: Optional[str]



from datetime import datetime
from typing import Optional

from pydantic import Field

from ..base import IDEX, DoubleNullableUomValue, IDEXAudit, IDEXAuditLite


class Maintenance(IDEXAuditLite):
    """Describes a maintenance event."""

    id: str
    location_id: Optional[str]
    type: Optional[str]
    description: Optional[str]
    repair_notes: Optional[str]
    running_meters: DoubleNullableUomValue
    start_date: datetime
    end_date: Optional[datetime]


class Run(IDEXAudit):
    """Describes a run."""

    id: str
    name: Optional[str]
    start_time: datetime
    end_time: Optional[datetime]
    status: str  # Enum but docs doesn't specify
    run_task: Optional[str]
    job_type: Optional[str]
    unit_id: Optional[str]
    hidden: bool
    wellbore_id: str
    log_ids: list[str]
    work_order_number: Optional[str] = Field(default=None)


class SurveyStation(IDEX):
    """Describes a single survey station."""

    id: str
    md: float
    tvd: Optional[float]
    incl: float
    azi: float
    dls: Optional[float]
    vert_sect: Optional[float]
    disp_ns: Optional[float]
    disp_ew: Optional[float]


class Survey(IDEX):
    """Describes a wellbore survey."""

    id: str
    name: Optional[str]
    wellbore_id: str
    md_uom: Optional[str]
    tvd_uom: Optional[str]
    inc_uom: Optional[str]
    azi_uom: Optional[str]
    dls_uom: Optional[str]
    vert_sect_uom: Optional[str]
    disp_ns_uom: Optional[str]
    disp_ew_uom: Optional[str]
    audit: IDEXAudit
    stations: list[SurveyStation]


class JobHistory(IDEXAuditLite):
    """Describes a historical job event."""

    id: str
    coiled_tubing_string: Optional[str]
    coiled_tubing_string_id: Optional[str]
    reel_name: Optional[str]
    reel_id: Optional[str]
    customer: Optional[str]
    customer_id: Optional[str]
    well_name: Optional[str]
    well_id: Optional[str]
    well_time_zone: Optional[str]
    added_dhrm: DoubleNullableUomValue = Field(alias="addedDHRM")
    weight_pull_avg: DoubleNullableUomValue = Field(alias="weightPullAverage")
    weight_pull_min: DoubleNullableUomValue
    weight_pull_max: DoubleNullableUomValue
    weight_push_avg: DoubleNullableUomValue = Field(alias="weightPushAverage")
    weight_push_min: DoubleNullableUomValue
    weight_push_max: DoubleNullableUomValue
    speed_pooh_avg: DoubleNullableUomValue = Field(alias="speedPOOHAverage")
    speed_pooh_min: DoubleNullableUomValue = Field(alias="speedPOOHMin")
    speed_pooh_max: DoubleNullableUomValue = Field(alias="speedPOOHMax")
    speed_rih_avg: DoubleNullableUomValue = Field(alias="speedRIHAverage")
    speed_rih_min: DoubleNullableUomValue = Field(alias="speedRIHMin")
    speed_rih_max: DoubleNullableUomValue = Field(alias="speedRIHMax")
    start_time: Optional[datetime]
    end_time: Optional[datetime]


class ScheduledJob(IDEXAuditLite):
    id: str
    name: Optional[str]
    job_plan: Optional[str]
    customer: Optional[str]
    customer_id: Optional[str]
    well_name: Optional[str]
    well_id: Optional[str]
    wellbore_name: Optional[str]
    wellbore_id: Optional[str]
    well_time_zone: Optional[str]
    start_time: Optional[datetime]
    end_time: Optional[datetime]


class UnitActiveWellbore(IDEXAudit):
    id: str
    unit_id: str
    wellbore_id: str


class WellboreHistory(IDEXAudit):
    id: str
    unit_id: Optional[str]
    wellbore_id: Optional[str]
    start: datetime
    end: Optional[datetime]
    log_ids: Optional[list[str]]



"""Lazy abstraction for HTTP client."""

from typing import Any
from urllib.parse import urljoin

from requests import Response, Session

from .auth import IDEXAuth


class IDEXApi:
    def __init__(
        self,
        *,
        auth: IDEXAuth,
        session: Session,
    ):
        self.auth = auth
        self.session = session

    @property
    def base_url(self):
        return self.auth.base_url

    def _authenticate(self) -> None:
        new_auth = self.auth.get_auth_header(base_url=self.base_url)
        self.session.headers.update(new_auth)

    def _send_request(self, *, method: str, url: str, **kwargs: Any) -> Response:
        self._authenticate()
        request_url = urljoin(self.base_url, url)
        rsp = self.session.request(method=method, url=request_url, **kwargs)
        rsp.raise_for_status()
        return rsp

    def get(self, *, url: str, **kwargs: Any) -> Response:
        return self._send_request(method="GET", url=url, **kwargs)

    def post(self, *, url: str, **kwargs: Any) -> Response:
        if "headers" in kwargs:
            headers = kwargs.pop("headers")
            if "Content-Type" not in headers:
                headers["Content-Type"] = "application/json"
            if "Accept" not in headers:
                headers["Accept"] = "application/json"
        else:
            headers = {"Content-Type": "application/json", "Accept": "application/json"}

        print(headers)
        print(kwargs.get("data"))
        return self._send_request(method="POST", url=url, headers=headers, **kwargs)



from datetime import timedelta
from typing import Optional

from pydantic import Field

from ..base import IDEX


class _TestBase(IDEX):
    """Common fields for all tests."""

    wellbore: Optional[str]
    start_time: Optional[str]
    end_time: Optional[str]
    sensor_name: Optional[str]
    created_date: Optional[str]
    comment: Optional[str]
    result: Optional[bool]
    user: Optional[str]


class _StartEndPressureTest(_TestBase):
    """Common fields for many pressure tests."""

    start_pressure_bar: Optional[float]
    end_pressure_bar: Optional[float]
    duration: timedelta


class BuildUpTest(_StartEndPressureTest):
    result_percentage: Optional[float]
    result_bar: Optional[float]


class InflowTest(BuildUpTest):
    """Describes a flow test: InflowTest."""

    delta_bar: Optional[float]


class Inflow30MinTest(InflowTest): ...


class PressureTest(_TestBase):
    """Describes a pressure test."""

    lp_start_pressure_bar: Optional[float]
    lp_end_pressure_bar: Optional[float]
    lp_duration: Optional[timedelta]
    lp_result_percentage: Optional[float]
    lp_result_bar: Optional[float]
    lp_reference_delta_p_bar: Optional[float]
    hp_start_pressure_bar: Optional[float]
    hp_end_pressure_bar: Optional[float]
    hp_duration: Optional[timedelta]
    hp_result_percentage: Optional[float]
    hp_result_bar: Optional[float]


class LeakRateTest(_StartEndPressureTest):
    """Describes an API leak rate test."""

    volume_m3: Optional[float]
    temperature_c: Optional[float]
    fluid_type: Optional[str]
    leak_rate_lpm: Optional[float] = Field(alias="leakRateLperMin")
    criteria_lpm: Optional[float] = Field(alias="criteriaLperMin")


class SssvLeakRateTest(LeakRateTest):
    """Describes a SSSV API leak rate test."""

    sssv_depth_meter: Optional[float]


class AsvLeakRateTest(LeakRateTest):
    """Describe an ASV API leak rate test."""

    asv_depth_meter: Optional[float]



from typing import Literal

from ....data_schemas.v1.events import (
    AsvLeakRateTest,
    BuildUpTest,
    Inflow30MinTest,
    InflowTest,
    LeakRateTest,
    PressureTest,
    SssvLeakRateTest,
)
from ..api import IDEXApi


class Tests:
    def __init__(self, api: IDEXApi) -> None:
        self._api = api

    def get_buildup_tests(self, kind: Literal["API", "QTS"]) -> list[BuildUpTest]:
        """
        Get `BuildUpTest` object(s).

        Returns
        -------
        list[BuildUpTest]

        """
        if kind == "API":
            url = "InterventionTests/ApiBuildUpTest"
        elif kind == "QTS":
            url = "InterventionTests/QtsTests"
        else:
            raise ValueError(f"Invalid kind: {kind}.")

        data = self._api.get(url=url)

        if data.status_code == 204:
            return []

        return [BuildUpTest.model_validate(row) for row in data.json()]

    def get_asv_leak_rate_tests(self) -> list[AsvLeakRateTest]:
        """
        Get `AsvLeakRateTest` object(s).

        Returns
        -------
        list[AsvLeakRateTest]

        """
        data = self._api.get(url="InterventionTests/AsvApiLeakRateTests")

        if data.status_code == 204:
            return []

        return [AsvLeakRateTest.model_validate(row) for row in data.json()]

    def get_sssv_leak_rate_tests(self) -> list[SssvLeakRateTest]:
        """
        Get `SssvLeakRateTest` object(s).

        Returns
        -------
        list[SssvLeakRateTest]

        """
        data = self._api.get(url="InterventionTests/SssvApiLeakRateTests")

        if data.status_code == 204:
            return []

        return [SssvLeakRateTest.model_validate(row) for row in data.json()]

    def get_inflow_tests(self, kind: Literal["30m", "Regular"]) -> list[InflowTest]:
        """
        Get `InflowTest` object(s).

        Parameters
        ----------
        kind : Literal["30m","Regular"]
            The kind of InflowTest to retrieve.

        Returns
        -------
        list[InflowTest]

        """
        if kind == "30m":
            url = "InterventionTests/Inflow30MinTests"
            model: type[InflowTest] = Inflow30MinTest
        elif kind == "Regular":
            url = "InterventionTests/InflowTests"
            model = InflowTest
        else:
            raise ValueError(f"Invalid kind: {kind}.")

        data = self._api.get(url=url)

        if data.status_code == 204:
            return []

        return [model.model_validate(row) for row in data.json()]

    def get_pressure_tests(self) -> list[PressureTest]:
        """
        Get `PressureTest` object(s).

        Returns
        -------
        list[PressureTest]

        """
        data = self._api.get(url="InterventionTests/PressureTests")

        if data.status_code == 204:
            return []

        return [PressureTest.model_validate(row) for row in data.json()]

    def get_xmt_leak_rate_tests(self, kind: Literal["HMV", "LMV"]) -> list[LeakRateTest]:
        """
        Get `PressureTest` object(s).

        Returns
        -------
        list[PressureTest]

        """
        if kind == "HMV":
            url = "InterventionTests/XmtHmvApiLeakRateTests"
        elif kind == "LMV":
            url = "InterventionTests/XmtLmvApiLeakRateTests"
        else:
            raise ValueError(f"Invalid kind: {kind}.")

        data = self._api.get(url=url)

        if data.status_code == 204:
            return []

        return [LeakRateTest.model_validate(row) for row in data.json()]



"""Contains Schemas for Physical assets."""

from datetime import datetime
from typing import Optional

from pydantic import Field as PField

from .base import IDEX, DoubleNullableUomValue, IDEXAudit, IDEXAuditLite


class Wellbore(IDEXAudit):
    id: str
    name: str
    well_id: Optional[str]
    item_state: Optional[str]


class Well(IDEXAudit):
    id: str
    name: Optional[str] = PField(default=None)
    facility: Optional[str] = PField(default=None)
    field: Optional[str] = PField(default=None)
    installation: Optional[str] = PField(default=None)
    item_state: Optional[str] = PField(default=None)
    latitude: Optional[float] = PField(default=None)
    longitude: Optional[float] = PField(default=None)
    operator: Optional[str] = PField(default=None)
    region: Optional[str] = PField(default=None)
    time_zone: Optional[str] = PField(default=None)
    well_number: Optional[str] = PField(default=None)
    reference_point: Optional[str] = PField(default=None)
    operator_id: Optional[str] = PField(default=None)
    field_id: Optional[str] = PField(default=None)
    installation_id: Optional[str] = PField(default=None)
    reference_point_elevation: Optional[DoubleNullableUomValue] = PField(default=None)


class WellboreLiveStatus(IDEXAudit):
    id: str
    wellbore_id: str
    is_live: bool
    is_live_last_changed: datetime
    is_depth_live: bool
    is_depth_live_last_changed: datetime
    created_by: Optional[str]
    modified_by: Optional[str]
    deleted_by: Optional[str]


class Unit(IDEX):
    id: str
    name: str
    type: str


class Equipment(IDEXAuditLite):
    id: str
    manufacturer: Optional[str]
    manufactured_date: Optional[datetime]
    next_maintenance_type: Optional[str]
    next_maintenance_alarm_type: Optional[str]
    next_maintenance_date: Optional[datetime]
    next_maintenance_days_left: DoubleNullableUomValue
    next_maintenance_meterage_left: DoubleNullableUomValue
    total_dhrm: DoubleNullableUomValue = PField(alias="totalDHRM")


class Reel(Equipment):
    core_diameter: DoubleNullableUomValue
    outer_diameter: DoubleNullableUomValue
    coiled_tubing_string: Optional[str]
    assigned_to: Optional[str]
    width: DoubleNullableUomValue
    coil_avg_life_consumption_percentage: DoubleNullableUomValue
    coil_max_life_consumption_percentage: DoubleNullableUomValue
    coil_max_life_consumption_position: DoubleNullableUomValue


class InjectorHead(Equipment):
    type: Optional[str]
    max_lifting_capacity: DoubleNullableUomValue
    total_assembled_weight: DoubleNullableUomValue
    calculate_dhrm_from_date: Optional[datetime] = PField(alias="calculateDHRMFromDate")
    initial_dhrm: DoubleNullableUomValue = PField(alias="initialDHRM")
    added_dhrm: DoubleNullableUomValue = PField(alias="addedDHRM")
    status: Optional[str]
    serial_number: Optional[str]
    gooseneck_radius: DoubleNullableUomValue
    assigned_to: Optional[str]
    district: Optional[str]


class Field(IDEXAudit):
    id: str
    name: Optional[str]


class Customer(IDEXAudit):
    id: str
    name: Optional[str]
    street_address: Optional[str]


class Log(IDEXAudit):
    id: str
    name: Optional[str]
    description: Optional[str]
    index_type: Optional[str]
    status: Optional[str]
    source_name: Optional[str]


class _Tangible(IDEX):
    type: Optional[str]
    od: DoubleNullableUomValue
    id: DoubleNullableUomValue
    weight_per_length: DoubleNullableUomValue = PField(alias="weightPrLength")
    md_top: DoubleNullableUomValue
    md_bottom: DoubleNullableUomValue
    abs_roughness: DoubleNullableUomValue
    comment: Optional[str]


class Casing(_Tangible): ...


class Completion(_Tangible):
    drift: DoubleNullableUomValue


class Deposit(IDEX):
    type: Optional[str]
    md_top: DoubleNullableUomValue
    md_bottom: DoubleNullableUomValue
    inner_diameter: DoubleNullableUomValue
    comment: Optional[str]
    label: Optional[str]
    relative_roughness: DoubleNullableUomValue


class Fish(IDEX):
    type: Optional[str]
    other_text: Optional[str]
    cable_type: Optional[str]
    bha_id: Optional[str]
    md_top: DoubleNullableUomValue
    outer_diameter: DoubleNullableUomValue
    inner_diameter: DoubleNullableUomValue = PField(alias="innerDiameters")
    weight: DoubleNullableUomValue
    comment: Optional[str]
    label: Optional[str]


class OpenHole(IDEX):
    hole_diameter: DoubleNullableUomValue
    md_top: DoubleNullableUomValue
    md_bottom: DoubleNullableUomValue
    abs_roughness: DoubleNullableUomValue
    comment: Optional[str]


class Coil(IDEX):
    id: str
    name: Optional[str]
    reel_id: Optional[str]
    reel_name: Optional[str]
    manufacturer: Optional[str]
    grade: Optional[str]
    unit_id: Optional[str]
    unit_name: Optional[str]
    comment: Optional[str]
    current_length: float
    original_length: float
    outer_diameter: float
    avg_fatigue: Optional[float] = PField(alias="averageFatigue")
    max_fatigue: float = PField(alias="maximumFatigue")
    max_fatigue_position: float = PField(alias="maximumFatiguePosition")
    fatigue_limit_warning: Optional[float] = PField(alias="fatigueLimitWarn")
    fatigue_limit_alarm: Optional[float]
    footage: Optional[float]
    date_manufactured: Optional[datetime]
    date_in_service: Optional[datetime]
    location_id: Optional[str]
    location_name: Optional[str]
    last_used: Optional[datetime]
    status: Optional[str]
    last_reel_id: Optional[str]
    last_reel_name: Optional[str]
    last_unit_id: Optional[str]
    last_unit_name: Optional[str]
    retire_reason: Optional[str]
    retire_comment: Optional[str]
    disposition: Optional[str]
    is_retired: bool



import logging
from typing import Any, Optional, overload

from ....data_schemas.v1.assets import Wellbore
from ....data_schemas.v1.events import Survey, SurveyStation
from ..api import IDEXApi


class Surveys:
    def __init__(self, api: IDEXApi) -> None:
        self._api = api

    @overload
    def get(
        self,
        *,
        filter: Optional[str] = None,
        select: Optional[list[str]] = None,
        top: Optional[int] = None,
        skip: Optional[int] = None,
        order_by: Optional[str] = None,
    ) -> list[Survey]: ...

    @overload
    def get(self, *, wellbore: Wellbore) -> list[Survey]: ...
    @overload
    def get(self, *, wellbore_id: str) -> list[Survey]: ...

    def get(
        self,
        *,
        wellbore: Optional[Wellbore] = None,
        wellbore_id: Optional[str] = None,
        filter: Optional[str] = None,
        select: Optional[list[str]] = None,
        top: Optional[int] = None,
        skip: Optional[int] = None,
        order_by: Optional[str] = None,
    ) -> list[Survey]:
        """
        Get `Survey` objects.

        Parameters
        ----------
        wellbore : Optional[Wellbore]
            Wellbore object to get Surveys for.
        wellbore_id : Optional[str]
            Wellbore ID to get Surveys for.
        filter : Optional[str]
            OData filter string.
        select : list[str] | None
            Provide a list of columns to retrieve from output.
        top : Optional[int]
            Limit the number of results returned.
        skip : Optional[int]
            Skip the first N results.
        order_by : Optional[str]
            Order the results by columns.

        Returns
        -------
        list[Survey]
            The `Survey` objects.

        """
        if wellbore is not None:
            # Get for singular wellbore
            logging.debug(f"Getting Surveys for Wellbore with ID: {wellbore.id}")
            data = self._api.get(url=f"Wellbores/{wellbore.id}/Surveys")

        elif wellbore_id is not None:
            # Get for singular wellbore
            logging.debug(f"Getting Surveys for Wellbore with ID: {wellbore_id}")
            data = self._api.get(url=f"Wellbores/{wellbore_id}/Surveys")

        else:
            # Get all surveys matching filters
            params: dict[str, Any] = {}
            if filter is not None:
                params["$filter"] = filter
            if select is not None:
                params["$select"] = ",".join(select)
            if top is not None:
                params["$top"] = top
            if skip is not None:
                params["$skip"] = skip
            if order_by is not None:
                params["$orderby"] = order_by

            data = self._api.get(url="Surveys", params=params)

        if data.status_code == 204:
            return []

        return [Survey.model_validate(row) for row in data.json()]

    def _get_stations(self, survey_id: str) -> list[SurveyStation]:
        data = self._api.get(url=f"Surveys/{survey_id}/Stations")
        if data.status_code == 204:
            return []
        return [SurveyStation.model_validate(row) for row in data.json()]

    @overload
    def get_stations(self, *, survey: Survey) -> list[SurveyStation]: ...
    @overload
    def get_stations(self, *, survey_id: str) -> list[SurveyStation]: ...

    def get_stations(
        self,
        *,
        survey: Optional[Survey] = None,
        survey_id: Optional[str] = None,
    ) -> list[SurveyStation]:
        if all(v is None for v in [survey, survey_id]):
            raise ValueError("Must provide either a Survey object or a survey_id.")

        if survey is not None:
            return self._get_stations(survey.id)
        elif survey_id is not None:
            return self._get_stations(survey_id)
        else:
            raise ValueError("Invalid input. Must provide either a Survey object or a survey_id.")



import logging
from typing import Any, Optional, Union, overload

from ....data_schemas.v1.assets import Well, Wellbore
from ..api import IDEXApi


class Wellbores:
    def __init__(self, api: IDEXApi) -> None:
        self._api = api

    @overload
    def get(self, *, id: str) -> Wellbore: ...

    @overload
    def get(
        self,
        *,
        filter: Optional[str] = None,
        select: Optional[list[str]] = None,
        top: Optional[int] = None,
        skip: Optional[int] = None,
        order_by: Optional[str] = None,
        include_soft_delete: Optional[bool] = False,
    ) -> list[Wellbore]: ...

    @overload
    def get(self, *, well: Well, include_soft_delete: Optional[bool] = False) -> list[Wellbore]: ...
    @overload
    def get(self, *, well_id: str, include_soft_delete: Optional[bool] = False) -> list[Wellbore]: ...

    def get(
        self,
        *,
        id: Optional[str] = None,
        well: Optional[Well] = None,
        well_id: Optional[str] = None,
        filter: Optional[str] = None,
        select: Optional[list[str]] = None,
        top: Optional[int] = None,
        skip: Optional[int] = None,
        order_by: Optional[str] = None,
        include_soft_delete: Optional[bool] = False,
    ) -> Union[Wellbore, list[Wellbore]]:
        """
        Get `Wellbore` object(s).

        Parameters
        ----------
        id : Optional[str]
            Well to retrieve.
        well : Optional[Well]
            Well object to get Wellbores for.
        well_id : Optional[str]
            Well ID to get Wellbores for.
        filter : Optional[str]
            OData filter string.
        select : list[str] | None
            Provide a list of columns to retrieve from output.
        top : Optional[int]
            Limit the number of results returned.
        skip : Optional[int]
            Skip the first N results.
        order_by : Optional[str]
            Order the results by columns.
        include_soft_delete : Optional[bool] = False
            Include soft deleted records.

        Returns
        -------
        Union[Wellbore, list[Wellbore]]
            The `Wellbore` object(s).

        """
        if id is not None:
            # Get singular well
            logging.debug("Getting Wellbore with ID: {id}")
            data = self._api.get(url=f"Wellbores/{id}")
            return Wellbore.model_validate(data.json())

        elif well is not None:
            logging.debug(f"Getting Wellbores for Well with ID: {well.id}")
            # Get Wellbores for singular well
            data = self._api.get(url=f"Wells/{well.id}/Wellbores")

        elif well_id is not None:
            # Get Wellbores for singular well
            logging.debug(f"Getting Wellbores for Well with ID: {well_id}")
            data = self._api.get(url=f"Wells/{well_id}/Wellbores")

        else:
            params: dict[str, Any] = {}
            if filter is not None:
                params["$filter"] = filter
            if select is not None:
                select = self._check_select(select)
                params["$select"] = ",".join(select)
            if top is not None:
                params["$top"] = top
            if skip is not None:
                params["$skip"] = skip
            if order_by is not None:
                params["$orderby"] = order_by

            data = self._api.get(url="Wellbores", params=params)

        if data.status_code == 204:
            return []

        wellbores = [Wellbore.model_validate(row) for row in data.json()]

        if include_soft_delete:
            return wellbores

        return [wb for wb in wellbores if wb.deleted_date is None]

    def _check_select(self, select: list[str]) -> list[str]:
        important_fields = ["id"]
        for field in important_fields:
            if field not in select:
                select.append(field)
        return select



from datetime import datetime, timedelta
from typing import Optional

from pydantic import BaseModel, ConfigDict, Field


def to_camel(string: str) -> str:
    """Convert snake_case string to camelCase."""
    string_split = string.split("_")
    return string_split[0] + "".join(word.capitalize() for word in string_split[1:])


class IDEX(BaseModel):
    """Base class for all IDEX data models."""

    model_config = ConfigDict(alias_generator=to_camel, populate_by_name=True)

    ...


class IDEXAuditLite(IDEX):
    """For subclasses with API audit fields."""

    created_date: Optional[datetime] = Field(default=None)
    modified_date: Optional[datetime] = Field(default=None)


class IDEXAudit(IDEXAuditLite):
    """For subclasses with API audit fields incl. deleted date."""

    deleted_date: Optional[datetime] = Field(default=None)


class TimeDuration(IDEX):
    """Describes a Time Duration."""

    type: Optional[str]
    duration: timedelta


class DoubleNullableUomValue(IDEX):
    """Describes a nullable value with unit."""

    value: Optional[float]
    uom: Optional[str]


class Approval(IDEX):
    """Describes an Approval."""

    name: Optional[str]
    approved: bool


class Change(IDEX):
    """Describes a Change in Changelog."""

    id: Optional[str]
    anchor: int
    type: Optional[str]
    source: Optional[str]
    created_date: datetime



import logging
from typing import Optional, overload

from ....data_schemas.v1.assets import Unit, Wellbore
from ....data_schemas.v1.events import UnitActiveWellbore
from ..api import IDEXApi
from .text_utils import url_encode_id


class UnitActiveWellbores:
    def __init__(self, api: IDEXApi) -> None:
        self._api = api

    @overload
    def get(self, *, include_soft_delete: Optional[bool] = False) -> list[UnitActiveWellbore]: ...
    @overload
    def get(self, *, unit: Unit, include_soft_delete: Optional[bool] = False) -> list[UnitActiveWellbore]: ...
    @overload
    def get(self, *, unit_id: str, include_soft_delete: Optional[bool] = False) -> list[UnitActiveWellbore]: ...
    @overload
    def get(self, *, wellbore: Wellbore, include_soft_delete: Optional[bool] = False) -> list[UnitActiveWellbore]: ...
    @overload
    def get(self, *, wellbore_id: str, include_soft_delete: Optional[bool] = False) -> list[UnitActiveWellbore]: ...

    def get(
        self,
        *,
        unit: Optional[Unit] = None,
        unit_id: Optional[str] = None,
        wellbore: Optional[Wellbore] = None,
        wellbore_id: Optional[str] = None,
        include_soft_delete: Optional[bool] = False,
    ) -> list[UnitActiveWellbore]:
        """
        Get `UnitActiveWellbore` object(s).

        Parameters
        ----------
        unit : Optional[Unit]
            Unit object to get UnitActiveWellbore for.
        unit_id : Optional[str]
            Unit ID to get UnitActiveWellbore for.
        wellbore : Optional[Wellbore]
            Wellbore object to get UnitActiveWellbore for.
        wellbore_id : Optional[str]
            Wellbore ID to get UnitActiveWellbore for.
        include_soft_delete : Optional[bool] = False
            Include soft deleted records.

        Returns
        -------
        list[UnitActiveWellbore]
            The `UnitActiveWellbore` object(s).

        """
        if unit is not None or unit_id is not None:
            id = unit_id if unit is None else unit.id
            assert id is not None
            logging.debug(f"Getting UnitActiveWellbores for Unit with ID: {id}")
            id = url_encode_id(id)
            data = self._api.get(url=f"UnitActiveWellbore/{id}")

        elif wellbore is not None or wellbore_id is not None:
            id = wellbore_id if wellbore is None else wellbore.id
            assert id is not None
            logging.debug(f"Getting UnitActiveWellbores for Wellbore with ID: {id}")
            id = url_encode_id(id)
            data = self._api.get(url=f"UnitActiveWellbore/{id}")

        else:
            data = self._api.get(url="UnitActiveWellbore")

        if data.status_code == 204:
            return []

        records = [UnitActiveWellbore.model_validate(row) for row in data.json()]

        if include_soft_delete:
            return records

        return [rec for rec in records if rec.deleted_date is None]



from datetime import datetime
from typing import Any, Optional

from pydantic import Field, field_validator

from .base import IDEX, IDEXAudit


class ChannelRange(IDEX):
    """Describes a Range."""

    channel_id: str = Field(alias="id")
    start: datetime
    end: datetime


class TimeRange(IDEX):
    """Describes a Time Range."""

    start: datetime
    end: datetime


class DataPoint(IDEX):
    time: datetime
    value: Any


class ChannelDataResponse(IDEX):
    id: str
    points: Optional[list[DataPoint]]


class _DataRequest(IDEX):
    limit: int
    ignore_unknown_ids: bool
    include_outside_points: bool

    @field_validator("limit")
    @classmethod
    def check_is_positive(cls, value: int):
        if int(value) <= 0:
            raise ValueError("limit must be positive")
        return value


class ChannelDataRequest(_DataRequest):
    ids: list[str]
    start: Optional[datetime]
    end: Optional[datetime]


class ChannelDataRangeRequest(_DataRequest):
    channels: list[ChannelRange]


class Channel(IDEXAudit):
    id: str
    name: Optional[str]
    global_name: Optional[str]
    description: Optional[str]
    uom: Optional[str]
    uom_class: Optional[str]
    data_type: Optional[str]
    index_type: Optional[str]
    status: Optional[str]
    range: TimeRange
    log_id: Optional[str] = Field(default=None)  # Added manually
    run_id: Optional[str] = Field(default=None)  # Added manually



from .events import (
    JobHistory,
    Maintenance,
    Run,
    ScheduledJob,
    Survey,
    SurveyStation,
    UnitActiveWellbore,
    WellboreHistory,
)
from .soe import SoeActivity, SoeChemicalMeasurement, SoeJob, SoeSensorValues, SoeTask
from .tests import (
    AsvLeakRateTest,
    BuildUpTest,
    Inflow30MinTest,
    InflowTest,
    LeakRateTest,
    PressureTest,
    SssvLeakRateTest,
)

__all__ = [
    # Events
    "JobHistory",
    "Maintenance",
    "Run",
    "ScheduledJob",
    "Survey",
    "SurveyStation",
    "UnitActiveWellbore",
    "WellboreHistory",
    # Soe
    "SoeActivity",
    "SoeChemicalMeasurement",
    "SoeJob",
    "SoeSensorValues",
    "SoeTask",
    # Tests
    "AsvLeakRateTest",
    "BuildUpTest",
    "InflowTest",
    "Inflow30MinTest",
    "LeakRateTest",
    "PressureTest",
    "SssvLeakRateTest",
]



"""Some helper utils relevant across modules."""

from urllib.parse import quote


def url_encode_id(id: str) -> str:
    """
    Encode a non-UUID ID for use in a URL.

    >>> id = "1/3-K-1"
    >>> assert url_encode_id(id) == "1%2F3-K-1"
    """
    return quote(id, safe="")
